#ifndef FS_UAE_OPTIONS_H
#define FS_UAE_OPTIONS_H

#define OPTION_ACCELERATOR "accelerator"
#define OPTION_ACCELERATOR_ROM "accelerator_rom"
#define OPTION_ACCELERATOR_MEMORY "accelerator_memory"
#define OPTION_ACCURACY "accuracy"
#define OPTION_AMIGA_MODEL "amiga_model"
#define OPTION_BLIZZARD_SCSI_KIT "blizzard_scsi_kit"
#define OPTION_BSDSOCKET_LIBRARY "bsdsocket_library"
#define OPTION_CDFS "cdfs"
#define OPTION_CHIP_MEMORY "chip_memory"
//#define OPTION_CURSOR_INTEGRATION "cursor_integration"
#define OPTION_CPU "cpu"
#define OPTION_CPU_IDLE "cpu_idle"
#define OPTION_DETERMINISTIC "deterministic"
#define OPTION_DONGLE_TYPE "dongle_type"
#define OPTION_FAST_MEMORY "fast_memory"
#define OPTION_FLOPPY_DRIVE_0 "floppy_drive_0"
#define OPTION_FLOPPY_DRIVE_VOLUME "floppy_drive_volume"
#define OPTION_FLOPPY_DRIVE_VOLUME_EMPTY "floppy_drive_volume_empty"
#define OPTION_FPU "fpu"
#define OPTION_GRAPHICS_CARD "graphics_card"
#define OPTION_GRAPHICS_CARD_ROM "graphics_card_rom"
#define OPTION_GRAPHICS_CARD_MEMORY "graphics_card_memory"
#define OPTION_JIT_COMPILER "jit_compiler"
#define OPTION_JIT_MEMORY "jit_memory"
#define OPTION_JOYSTICK_PORT_0_AUTOSWITCH "joystick_port_0_autoswitch"
#define OPTION_LINE_DOUBLING "line_doubling"
#define OPTION_LOG_AUTOSCALE "log_autoscale"
#define OPTION_LOG_BSDSOCKET "log_bsdsocket"
#define OPTION_LOW_RESOLUTION "low_resolution"
#define OPTION_MOTHERBOARD_RAM "motherboard_ram"
#define OPTION_MMU "mmu"
#define OPTION_MODEL "model"
#define OPTION_SERIAL_PORT "serial_port"
#define OPTION_SLOW_MEMORY "slow_memory"
#define OPTION_SOUND_CARD "sound_card"
#define OPTION_STEREO_SEPARATION "stereo_separation"
#define OPTION_THEME_ZOOM "theme_zoom"
#define OPTION_UAENATIVE_LIBRARY "uaenative_library"
#define OPTION_WORKBENCH_DISK "workbench_disk"
#define OPTION_ZORRO_III_MEMORY "zorro_iii_memory"

/* Deprecated options */

#define OPTION_UAEGFX_CARD "uaegfx_card"

#endif /* FS_UAE_OPTIONS_H */
