#ifndef FS_EMU_OPTIONS_H
#define FS_EMU_OPTIONS_H

#define OPTION_AUDIO_DRIVER "audio_driver"
#define OPTION_SCREENSHOTS_OUTPUT_DIR "screenshots_output_dir"
#define OPTION_VIDEO_DRIVER "video_driver"

#include <fs/ml/options.h>

#endif /* FS_EMU_OPTIONS_H */
