#ifndef FS_EMU_BENCHMARK_H
#define FS_EMU_BENCHMARK_H

extern bool g_fs_emu_benchmark_mode;
extern volatile int g_fs_emu_benchmark_emu_frame_count;
extern volatile int g_fs_emu_benchmark_sys_frame_count;

#endif // FS_EMU_BENCHMARK_H
