#ifndef FS_FS_H_
#define FS_FS_H_

#include <fs/defines.h>
#include <fs/base.h>
#include <fs/init.h>
#include <fs/filesys.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __cplusplus
}
#endif

#endif // FS_FS_H_
