#ifndef UAE_CYCLEUNIT_H
#define UAE_CYCLEUNIT_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

// FIXME: this is temporarily put in a separate header file,
// should be merged into events.h
#define CYCLE_UNIT 512
#define OFFICIAL_CYCLE_UNIT 512

#endif // UAE_CYCLEUNIT_H
